/*
 * kexec: Linux boots Linux
 *
 * Copyright (C) 2003,2004  Eric Biederman (ebiederm@xmission.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 2 of the License).
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>
#include <getopt.h>
#include <elf.h>
#include <boot/elf_boot.h>
#include <ip_checksum.h>
#include <x86/x86-linux.h>
#include "kexec.h"
#include "kexec-x86.h"


#define BOOTLOADER "kexec"
#define BOOTLOADER_VERSION VERSION
#define MAX_COMMAND_LINE 256

#define UPSZ(X) ((sizeof(X) + 3) &~3)
static struct boot_notes {
	Elf_Bhdr hdr;
	Elf_Nhdr bl_hdr;
	unsigned char bl_desc[UPSZ(BOOTLOADER)];
	Elf_Nhdr blv_hdr;
	unsigned char blv_desc[UPSZ(BOOTLOADER_VERSION)];
	Elf_Nhdr cmd_hdr;
	unsigned char command_line[0];
} 
elf_boot_notes = {
	.hdr = {
		.b_signature = 0x0E1FB007,
		.b_size = sizeof(elf_boot_notes),
		.b_checksum = 0,
		.b_records = 3,
	},
	.bl_hdr = {
		.n_namesz = 0,
		.n_descsz = sizeof(BOOTLOADER),
		.n_type = EBN_BOOTLOADER_NAME,
	},
	.bl_desc = BOOTLOADER,
	.blv_hdr = {
		.n_namesz = 0,
		.n_descsz = sizeof(BOOTLOADER_VERSION),
		.n_type = EBN_BOOTLOADER_VERSION,
	},
	.blv_desc = BOOTLOADER_VERSION,
	.cmd_hdr = {
		.n_namesz = 0,
		.n_descsz = 0,
		.n_type = EBN_COMMAND_LINE,
	},
};


static inline unsigned long align(unsigned long val, unsigned long align)
{
	return (val + align - 1) & ~(align - 1);

}
static int setup_linux_parameters(struct x86_linux_param_header *real_mode)
{
	struct memory_range *range;
	int i, ranges;
	
	/* Default screen size */
	real_mode->orig_x = 0;
	real_mode->orig_y = 0;
	real_mode->orig_video_page = 0;
	real_mode->orig_video_mode = 0;
	real_mode->orig_video_cols = 80;
	real_mode->orig_video_lines = 25;
	real_mode->orig_video_ega_bx = 0;
	real_mode->orig_video_isVGA = 1;
	real_mode->orig_video_points = 16;

	/* Fill in the memsize later */
	real_mode->ext_mem_k = 0;
	real_mode->alt_mem_k = 0;
	real_mode->e820_map_nr = 0;

	/* Default APM info */
	memset(&real_mode->apm_bios_info, 0, sizeof(real_mode->apm_bios_info));
	/* Default drive info */
	memset(&real_mode->drive_info, 0, sizeof(real_mode->drive_info));
	/* Default sysdesc table */
	real_mode->sys_desc_table.length = 0;

	/* default yes: this can be overridden on the command line */
	real_mode->mount_root_rdonly = 0xFFFF;

	/* default /dev/hda
	 * this can be overrident on the command line if necessary.
	 */
	real_mode->root_dev = (0x3 <<8)| 0;

	/* another safe default */
	real_mode->aux_device_info = 0;

	/* Fill in the memory info */
	if ((get_memory_ranges(&range, &ranges) < 0) || ranges == 0) {
		fprintf(stderr, "Cannot get memory information\n");
		return -1;
	}
	if (ranges > E820MAX) {
		fprintf(stderr, "Too many memory ranges, truncating...\n");
		ranges = E820MAX;
	}
	real_mode->e820_map_nr = ranges;
	for(i = 0; i < ranges; i++) {
		real_mode->e820_map[i].addr = range[i].start;
		real_mode->e820_map[i].size = range[i].end - range[i].start;
		switch (range[i].type) {
		case RANGE_RAM:
			real_mode->e820_map[i].type = E820_RAM; 
			break;
		case RANGE_ACPI:
			real_mode->e820_map[i].type = E820_ACPI; 
			break;
		case RANGE_ACPI_NVS:
			real_mode->e820_map[i].type = E820_NVS;
			break;
		default:
		case RANGE_RESERVED:
			real_mode->e820_map[i].type = E820_RESERVED; 
			break;
		}
		if (range[i].type != RANGE_RAM)
			continue;
		if ((range[i].start <= 0x100000) && range[i].end > 0x100000) {
			unsigned long long mem_k = (range[i].end >> 10) - 0x100000;
			real_mode->ext_mem_k = mem_k;
			real_mode->alt_mem_k = mem_k;
			if (mem_k > 0xfc00) {
				real_mode->ext_mem_k = 0xfc00; /* 64M */
			}
			if (mem_k > 0xffffffff) {
				real_mode->alt_mem_k = 0xffffffff;
			}
		}
	}
	/* Boot block magic */
	memcpy(real_mode->header_magic, "HdrS", 4);
	real_mode->protocol_version = 0x0202;
	
	/* Say I'm a boot loader */
	real_mode->loader_type = LOADER_TYPE_UNKNOWN;

	/* No loader flags */
	real_mode->loader_flags = 0;

	/* Ramdisk address and size */
	real_mode->initrd_start = 0;
	real_mode->initrd_size = 0;

	return 0;
}

int elf32_x86_probe(FILE *file)
{
	Elf32_Ehdr ehdr;
	if (fseek(file, 0, SEEK_SET) < 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	if (fread(&ehdr, sizeof(ehdr), 1, file)  != 1) {
		fprintf(stderr, "read error: %s\n",
			strerror(errno));
		return -1;
	}
	if (memcmp(ehdr.e_ident, ELFMAG, SELFMAG) != 0) {
		/* No ELF header */
		return 0;
	}
	if (ehdr.e_ident[EI_CLASS] != ELFCLASS32) {
		/* Not a 32bit ELF file */
		return 0;
	}
	if (ehdr.e_ident[EI_DATA] != ELFDATA2LSB) {
		/* not a little endian ELF file */
		return 0;
	}
	if ((ehdr.e_ident[EI_VERSION] != EV_CURRENT) ||
		(ehdr.e_version != EV_CURRENT)) {
		/* unkown elf version */
		return 0;
	}
	if (ehdr.e_type != ET_EXEC) {
		/* not an ELF executable */
		return 0;
	}

	if (ehdr.e_ehsize != sizeof(Elf32_Ehdr)) {
		/* invalid ELF header size */
		return 0;
	}
	if (ehdr.e_phentsize != sizeof(Elf32_Phdr)) {
		/* invalid program header size */
		return 0;
	}
	if ((ehdr.e_phoff == 0) || (ehdr.e_phnum == 0)) {
		/* no program header */
		return 0;
	}
	/* Verify the architecuture specific bits */
	if ((ehdr.e_machine != EM_386) && (ehdr.e_machine != EM_486)) {
		/* for a different architecture */
		return 0;
	}
	return 1;
}

void elf32_x86_usage(void)
{
	printf(	"    --command-line=STRING Set the kernel command line to STRING\n"
		"    --append=STRING       Set the kernel command line to STRING\n"
		"    --initrd=FILE         Use FILE as the kernel's initial ramdisk.\n"
		"    --ramdisk=FILE        Use FILE as the kernel's initial ramdisk.\n"
		"    --args-linux          Pass linux kernel style options\n"
		"    --args-elf            Pass elf boot notes\n"
		);
	
	
}

int elf32_x86_load(FILE *file, int argc, char **argv,
	void **ret_entry, struct kexec_segment **ret_segments, int *ret_nr_segments)
{
	struct kexec_segment *segment;
	int nr_segments;
	Elf32_Ehdr ehdr;
	Elf32_Phdr *phdr;
	size_t phdr_bytes;
	size_t arg_bytes, arg_align;
	struct boot_notes *notes;
	size_t note_bytes;
	const char *command_line;
	int command_line_len;
	const char *ramdisk;
	FILE *fp_ramdisk = 0;
	unsigned long ramdisk_length;
	int arg_style;
#define ARG_STYLE_ELF   0
#define ARG_STYLE_LINUX 1
	int i;
	int opt;
#define OPT_APPEND	OPT_MAX+0
#define OPT_RAMDISK	OPT_MAX+1
#define OPT_ARGS_ELF    OPT_MAX+2
#define OPT_ARGS_LINUX  OPT_MAX+3

	static const struct option options[] = {
		KEXEC_OPTIONS
		{ "command-line",	1, NULL, OPT_APPEND },
		{ "append",		1, NULL, OPT_APPEND },
		{ "initrd",		1, NULL, OPT_RAMDISK },
		{ "ramdisk",		1, NULL, OPT_RAMDISK },
		{ "args-elf",		0, NULL, OPT_ARGS_ELF },
		{ "args-linux",		0, NULL, OPT_ARGS_LINUX },
		{ 0, 			0, NULL, 0 },
	};

	static const char short_options[] = KEXEC_OPT_STR "";

	/*
	 * Parse the command line arguments
	 */
	arg_style = ARG_STYLE_ELF;
	command_line = 0;
	ramdisk = 0;
	ramdisk_length = 0;
	while((opt = getopt_long(argc, argv, short_options, options, 0)) != -1) {
		switch(opt) {
		default:
			/* Ignore core options */
			if (opt < OPT_MAX) {
				break;
			}
		case '?':
			usage();
			return -1;
		case OPT_APPEND:
			command_line = optarg;
			break;
		case OPT_RAMDISK:
			ramdisk = optarg;
			break;
		case OPT_ARGS_ELF: 
			arg_style = ARG_STYLE_ELF;
			break;
		case OPT_ARGS_LINUX:
			arg_style = ARG_STYLE_LINUX;
			break;
		}
	}
	command_line_len = 0;
	if (command_line) {
		command_line_len = strlen(command_line) +1;
	}

	if (ramdisk) {
		fp_ramdisk = fopen(ramdisk, "rb");
		if (!fp_ramdisk) {
			fprintf(stderr, "Cannot open %s:%s\n",
				ramdisk, strerror(errno));
			return -1;
		}
		if (fseek(fp_ramdisk, 0, SEEK_END) < 0) {
			fprintf(stderr, "seek error on %s: %s\n",
				ramdisk, strerror(errno));
			return -1;
		}
		ramdisk_length = ftell(fp_ramdisk);
	}


	/* Read in the Elf header */
	if (fseek(file, 0, SEEK_SET) != 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	if (fread(&ehdr, sizeof(ehdr), 1, file) != 1) {
		fprintf(stderr, "read error: %s\n",
			strerror(errno));
		return -1;
	}
	/* Read in the program header */
	phdr_bytes = sizeof(*phdr) *ehdr.e_phnum;
	phdr = malloc(phdr_bytes);
	if (phdr == 0) {
		fprintf(stderr, "malloc failed: %s\n",
			strerror(errno));
		return -1;
	}
	if (fseek(file, ehdr.e_phoff, SEEK_SET) != 0) {
		fprintf(stderr, "seek error: %s\n",
			strerror(errno));
		return -1;
	}
	if (fread(phdr, phdr_bytes, 1, file) != 1) {
		fprintf(stderr, "read error: %s\n",
			strerror(errno));
		return -1;
	}

	/* Setup the segments */
	segment = malloc(sizeof(*segment) * (ehdr.e_phnum +1));
	if (segment == 0) {
		fprintf(stderr, "malloc failed: %s\n",
			strerror(errno));
		return -1;
	}

	/* Skip the argument segment */
	nr_segments = 0;
	segment[nr_segments].buf = 0;
	segment[nr_segments].bufsz = 0;
	segment[nr_segments].mem = 0;
	segment[nr_segments].memsz = 0;
	nr_segments++;

	/* Now all the rest of the segments */
	for(i = 0; i < ehdr.e_phnum; i++) {
		char *buf;
		size_t size;
		if (phdr[i].p_type != PT_LOAD) {
			continue;
		}
		size = phdr[i].p_filesz;
		if (size > phdr[i].p_memsz) {
			size = phdr[i].p_memsz;
		}
		buf = malloc(size);
		if (buf == 0) {
			fprintf(stderr, "malloc failed: %s\n",
				strerror(errno));
			return -1;
		}
		segment[nr_segments].buf = buf;
		segment[nr_segments].bufsz = size;
		segment[nr_segments].mem = (void *)(unsigned long)phdr[i].p_paddr;
		segment[nr_segments].memsz = phdr[i].p_memsz;
		if (valid_memory_range(segment + nr_segments) < 0) {
			fprintf(stderr, "Invalid memory segment %p - %p\n",
				segment[nr_segments].mem,
				((char*)segment[nr_segments].mem) + segment[nr_segments].memsz);
			return -1;
		}
		nr_segments++;
		if (size == 0) {
			/* Don't do file I/O if there is nothing in the file */
			continue;
		}
		if (fseek(file, phdr[i].p_offset, SEEK_SET) != 0) {
			fprintf(stderr, "seek failed: %s\n",
				strerror(errno));
			return -1;
		}
		if (fread(buf, size, 1, file) != 1) {
			fprintf(stderr, "Read failed: %s\n",
				strerror(errno));
			return -1;
		}
	}
	/* Generate and setup the argument segment */
	if (sort_segments(segment, nr_segments) < 0) {
		return -1;
	}
	note_bytes = 0; /* Shutup gcc */
	if (arg_style == ARG_STYLE_ELF) {
		arg_align = 4;
		note_bytes = sizeof(elf_boot_notes) + 
			align(command_line_len, arg_align);
		arg_bytes = note_bytes + align(setup32_size, arg_align);
	}
	else if (arg_style == ARG_STYLE_LINUX) {
		arg_align = 65536;
		arg_bytes = align(setup32_size, arg_align) + 
			sizeof(struct x86_linux_faked_param_header);
	}
	else {
		fprintf(stderr, "Uknown argument style");
		return -1;
	}
	segment[0].bufsz = arg_bytes;
	segment[0].memsz = arg_bytes;
	if (locate_arguments(segment, nr_segments, arg_align, 0xFFFFFFFFUL) < 0) {
		return -1;
	}
	segment[0].buf = malloc(arg_bytes);
	if (segment[0].buf == 0) {
		fprintf(stderr, "malloc failed: %s\n",
			strerror(errno));
		return -1;
	}
	if (arg_style == ARG_STYLE_ELF) {
		notes = (struct boot_notes *)(((char *)segment[0].buf) + 
			align(setup32_size, arg_align));
		setup32_regs.eax = 0x0E1FB007;
		setup32_regs.ebx = ((unsigned long)segment[0].mem) + 
			align(setup32_size, arg_align);
		setup32_regs.eip = ehdr.e_entry;
		memcpy(segment[0].buf, setup32_start, setup32_size);
		memcpy(notes, &elf_boot_notes, sizeof(elf_boot_notes));
		memcpy(notes->command_line, command_line, command_line_len);
		notes->hdr.b_size = note_bytes;
		notes->cmd_hdr.n_descsz = command_line_len;
		notes->hdr.b_checksum = compute_ip_checksum(notes, note_bytes);

		/* Error if we have been passed a ramdisk */
		if (fp_ramdisk) {
			fprintf(stderr, "Ramdisks not supported with generic elf arguments");
			return -1;
		}
	}
	else if (arg_style == ARG_STYLE_LINUX) {
		struct x86_linux_faked_param_header *hdr;
		hdr = (struct x86_linux_faked_param_header *)
			(((char *)segment[0].buf) + align(setup32_size, arg_align));
		/* Clear the unused registers */
		setup32_regs.eax = 0;
		setup32_regs.ecx = 0;
		setup32_regs.edx = 0;
		setup32_regs.edi = 0;
		setup32_regs.esp = 0;
		setup32_regs.ebp = 0;
		
		/* Flag to indicate we are the bootstrap processor */
		setup32_regs.ebx = 0;

		/* Pointer to the parameters */
		setup32_regs.esi = ((unsigned long)segment[0].mem) + 
			align(setup32_size, arg_align);

		/* Set the entry point */
		setup32_regs.eip = ehdr.e_entry;

		/* Initialize the parameter header to 0 */
		memset(hdr, 0, sizeof(*hdr));
		
		/* Fill in the linux parameters */
		setup_linux_parameters(&hdr->hdr);

		/* Fill in the command line */
		if (command_line_len >= COMMAND_LINE_SIZE) {
			command_line_len = COMMAND_LINE_SIZE -1;
		}
		memcpy(hdr->command_line, command_line, command_line_len);
		hdr->command_line[command_line_len] = '\0';

		/* Record where I am putting the command line */
		hdr->hdr.cmd_line_ptr = setup32_regs.esi +
			offsetof(struct x86_linux_faked_param_header, command_line);
		
		/* Copy in the setup32 code */
		memcpy(segment[0].buf, setup32_start, setup32_size);

		/* Error if we have been passed a ramdisk */
		if (fp_ramdisk) {
			fprintf(stderr, "Ramdisks not yet supported with linux style arguments");
			return -1;
		}
		
	}

	*ret_entry = segment[0].mem;
	*ret_nr_segments = nr_segments;
	*ret_segments = segment;
	return 0;
}
